package io.gitlab.jfronny.inceptum.launcher.system.mds;

import io.gitlab.jfronny.commons.io.JFiles;
import io.gitlab.jfronny.commons.throwable.ThrowingConsumer;
import io.gitlab.jfronny.gson.JsonParseException;
import io.gitlab.jfronny.inceptum.common.Utils;
import io.gitlab.jfronny.inceptum.launcher.model.fabric.FabricModJson;
import io.gitlab.jfronny.inceptum.launcher.model.inceptum.ModDescription;
import io.gitlab.jfronny.inceptum.launcher.util.ModPath;
import io.gitlab.jfronny.inceptum.launcher.system.source.ModSource;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URISyntaxException;
import java.nio.file.*;
import java.util.List;
import java.util.Optional;
import java.util.function.BiConsumer;

public record FileScanTask(Path file, BiConsumer<Path, IWModDescription> discovered,
                           String gameVersion) implements Runnable {
    @Override
    public void run() {
        if (!Files.exists(file)) return;
        if (Files.isDirectory(file)) {
            discovered.accept(file, new IWModDescription(file));
            return;
        }
        try {
            if (ModPath.isJar(file)) {
                final var imod = new Object() {
                    Path i = file.getParent().resolve(file.getFileName() + ModPath.EXT_IMOD);
                    ModDescription md;
                };
                // mod description
                if (!Files.exists(imod.i)) {
                    JFiles.writeObject(imod.i, ModDescription.of(file));
                }
                imod.md = JFiles.readObject(imod.i, ModDescription.class);
                evaluateSources(imod.md, file, imod.i, newImod -> {
                    imod.i = newImod;
                    imod.md = JFiles.readObject(imod.i, ModDescription.class);
                });
                discovered.accept(imod.i, new IWModDescription(file, Optional.of(imod.md), getFmj(file, imod.md), Optional.of(imod.i)));
                return;
            }
            if (ModPath.isImod(file)) {
                String fn = file.getFileName().toString();
                Path modFile = file.getParent().resolve(fn.substring(0, fn.length() - 5));
                final var imod = new Object() {
                    Path i = file;
                    ModDescription md = JFiles.readObject(file, ModDescription.class);
                };
                evaluateSources(imod.md, modFile, imod.i, newImod -> {
                    imod.i = newImod;
                    imod.md = JFiles.readObject(imod.i, ModDescription.class);
                });
                discovered.accept(imod.i, new IWModDescription(imod.i, Optional.of(imod.md), getFmj(modFile, imod.md), Optional.of(imod.i)));
                return;
            }
            discovered.accept(file, new IWModDescription(file));
        } catch (IOException | URISyntaxException | JsonParseException e) {
            Utils.LOGGER.error("Could not scan file for mod info", e);
        }
    }

    private <TEx extends Throwable> void evaluateSources(ModDescription md, Path modFile, Path imodFile, ThrowingConsumer<Path, TEx> imodModified) throws IOException, TEx {
        boolean modified = false;
        if (md.initialize(gameVersion)) {
            JFiles.writeObject(imodFile, md);
            modified = true;
        }
        ModSource selectedSource = null;
        for (ModSource source : md.sources.keySet()) {
            source.getUpdate(gameVersion);
            if (!Files.exists(source.getJarPath())) source.download();
            selectedSource = source;
        }
        if (selectedSource != null && Files.exists(modFile)) {
            Files.delete(modFile);
            Path newImod = imodFile.getParent().resolve(selectedSource.getShortName() + ModPath.EXT_IMOD);
            Files.move(imodFile, newImod);
            imodFile = newImod;
            modified = true;
        }
        if (modified) imodModified.accept(imodFile);
    }

    private Optional<FabricModJson> getFmj(Path modJarDefault, ModDescription md) throws IOException, URISyntaxException {
        if (!Files.exists(modJarDefault)) {
            if (md.sources.isEmpty()) {
                throw new FileNotFoundException("Mod " + modJarDefault.getFileName().toString() + " doesn't specify a source and has no file");
            }
            modJarDefault = List.copyOf(md.sources.keySet()).get(0).getJarPath();
        }
        try (FileSystem fs = Utils.openZipFile(modJarDefault, false)) {
            return Files.exists(fs.getPath("fabric.mod.json"))
                    ? Optional.of(JFiles.readObject(fs.getPath("fabric.mod.json"), FabricModJson.class))
                    : Optional.empty();
        }
    }
}
